@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2019-2019. All Rights Reserved.
+%% Copyright Ericsson AB 2019-2021. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -98,8 +98,10 @@
          unsupported_sign_algo_cert_client_auth/1,
          longer_chain/0,
          longer_chain/1,
-         duplicate_chain/0,
-         duplicate_chain/1,
+         cross_signed_chain/0,
+         cross_signed_chain/1,
+         expired_root_with_cross_signed_root/0,
+         expired_root_with_cross_signed_root/1,
          key_auth_ext_sign_only/0,
          key_auth_ext_sign_only/1,
          hello_retry_request/0,
@@ -113,7 +115,13 @@
          hello_retry_client_auth_empty_cert_rejected/0,
          hello_retry_client_auth_empty_cert_rejected/1,
          basic_rsa_1024/0,
-         basic_rsa_1024/1
+         basic_rsa_1024/1,
+         signature_algorithms_bad_curve_secp256r1/0,
+         signature_algorithms_bad_curve_secp256r1/1,
+         signature_algorithms_bad_curve_secp384r1/0,
+         signature_algorithms_bad_curve_secp384r1/1,
+         signature_algorithms_bad_curve_secp521r1/0,
+         signature_algorithms_bad_curve_secp521r1/1
          ]).
 
 %%--------------------------------------------------------------------
@@ -133,21 +141,24 @@ all() ->
 groups() ->
     [
      {'tlsv1.3', [], tls_1_3_protocol_groups()}, 
-     {'tlsv1.2', [], tls_1_2_protocol_groups() -- [{group,rsa_pss_pss}]},
+     {'tlsv1.2', [], tls_1_2_protocol_groups()},
      {'tlsv1.1', [], ssl_protocol_groups()},
      {'tlsv1', [], ssl_protocol_groups()},
-     {'dtlsv1.2', [], tls_1_2_protocol_groups() -- [{group,rsa_pss_rsae}, {group,rsa_pss_pss}]},
+     {'dtlsv1.2', [], tls_1_2_protocol_groups()},
      {'dtlsv1', [], ssl_protocol_groups()},
-     {rsa, [], all_version_tests() ++ rsa_tests() ++ pre_tls_1_3_rsa_tests()},
+     {rsa, [], all_version_tests() ++ rsa_tests() ++ pre_tls_1_3_rsa_tests() ++ [client_auth_seelfsigned_peer]},
      {ecdsa, [], all_version_tests()},
      {dsa, [], all_version_tests()},
      {rsa_1_3, [], all_version_tests() ++ rsa_tests() ++
-          tls_1_3_tests() ++ tls_1_3_rsa_tests() ++ [basic_rsa_1024]},
-     {rsa_pss_rsae, [], all_version_tests() ++ rsa_tests()},
+          tls_1_3_tests() ++ tls_1_3_rsa_tests() ++ [client_auth_seelfsigned_peer, basic_rsa_1024]},
+     {rsa_pss_rsae, [], all_version_tests() ++ tls_1_2_rsa_tests()},
      {rsa_pss_rsae_1_3, [], all_version_tests() ++ rsa_tests() ++ tls_1_3_tests() ++ tls_1_3_rsa_tests()},
-     {rsa_pss_pss, [], all_version_tests() ++ rsa_tests()},
+     {rsa_pss_pss, [], all_version_tests()},
      {rsa_pss_pss_1_3, [], all_version_tests() ++ rsa_tests() ++ tls_1_3_tests() ++ tls_1_3_rsa_tests()},
-     {ecdsa_1_3, [], all_version_tests() ++ tls_1_3_tests()},
+     {ecdsa_1_3, [], all_version_tests() ++ tls_1_3_tests() ++
+          [signature_algorithms_bad_curve_secp256r1,
+           signature_algorithms_bad_curve_secp384r1,
+           signature_algorithms_bad_curve_secp521r1]},
      {eddsa_1_3, [], all_version_tests() ++ tls_1_3_tests()}
     ].
 
@@ -187,7 +198,9 @@ pre_tls_1_3_rsa_tests() ->
 
 rsa_tests() ->
    [
-    longer_chain
+    longer_chain,
+    cross_signed_chain,
+    expired_root_with_cross_signed_root
    ].
 
 tls_1_3_rsa_tests() ->
@@ -196,6 +209,12 @@ tls_1_3_rsa_tests() ->
       unsupported_sign_algo_cert_client_auth
      ].
 
+tls_1_2_rsa_tests() ->
+     [
+      unsupported_sign_algo_client_auth,
+      unsupported_sign_algo_cert_client_auth
+     ].
+
 all_version_tests() ->
     [
      no_auth,
@@ -207,7 +226,6 @@ all_version_tests() ->
      client_auth_do_not_allow_partial_chain,
      client_auth_partial_chain_fun_fail,
      client_auth_sni,
-     client_auth_seelfsigned_peer,
      missing_root_cert_no_auth,
      missing_root_cert_auth,
      missing_root_cert_auth_user_verify_fun_accept,
@@ -266,11 +284,30 @@ do_init_per_group(Group, Config0) when Group == rsa;
                    {server_cert_opts, SOpts} | 
                    lists:delete(server_cert_opts, 
                                 lists:delete(client_cert_opts, Config))])];
-
 do_init_per_group(Alg, Config) when Alg == rsa_pss_rsae;
-                                 Alg == rsa_pss_pss;
-                                 Alg == rsa_pss_rsae_1_3;
-                                 Alg == rsa_pss_pss_1_3 ->
+                                    Alg == rsa_pss_pss ->
+    Supports = crypto:supports(),
+    RSAOpts = proplists:get_value(rsa_opts, Supports),
+
+    case lists:member(rsa_pkcs1_pss_padding, RSAOpts)
+        andalso lists:member(rsa_pss_saltlen, RSAOpts)
+        andalso lists:member(rsa_mgf1_md, RSAOpts) of
+        true ->
+            #{client_config := COpts,
+              server_config := SOpts} = ssl_test_lib:make_rsa_pss_pem(rsa_alg(Alg), [], Config, ""),
+            [{cert_key_alg, Alg},
+             {extra_client, sig_algs(Alg)},
+             {extra_server, sig_algs(Alg)} |
+             lists:delete(cert_key_alg,
+                          [{client_cert_opts, COpts},
+                           {server_cert_opts, SOpts} |
+                           lists:delete(server_cert_opts,
+                                        lists:delete(client_cert_opts, Config))])];
+        false ->
+            {skip, "Missing EC crypto support"}
+    end;
+do_init_per_group(Alg, Config) when Alg == rsa_pss_rsae_1_3;
+                                    Alg == rsa_pss_pss_1_3 ->
 
     Supports = crypto:supports(),
     RSAOpts = proplists:get_value(rsa_opts, Supports),
@@ -356,14 +393,38 @@ do_init_per_group(_Group, Config) ->
 end_per_group(GroupName, Config) ->
   ssl_test_lib:end_per_group(GroupName, Config).
 
+init_per_testcase(signature_algorithms_bad_curve_secp256r1, Config) ->
+    init_rsa_ecdsa_opts(Config, secp256r1);
+init_per_testcase(signature_algorithms_bad_curve_secp384r1, Config) ->
+    init_rsa_ecdsa_opts(Config, secp384r1);
+init_per_testcase(signature_algorithms_bad_curve_secp521r1, Config) ->
+    init_rsa_ecdsa_opts(Config, secp521r1);
 init_per_testcase(_TestCase, Config) ->
     ssl_test_lib:ct_log_supported_protocol_versions(Config),
     ct:timetrap({seconds, 10}),
     Config.
 
-end_per_testcase(_TestCase, Config) ->     
+end_per_testcase(_TestCase, Config) ->
     Config.
 
+init_rsa_ecdsa_opts(Config0, Curve) ->
+    PKAlg = crypto:supports(public_keys),
+    case lists:member(ecdsa, PKAlg) andalso (lists:member(ecdh, PKAlg) orelse lists:member(dh, PKAlg)) of
+        true ->
+            Config = ssl_test_lib:make_rsa_ecdsa_cert(Config0, Curve),
+            COpts = proplists:get_value(client_rsa_ecdsa_opts, Config),
+            SOpts = proplists:get_value(server_rsa_ecdsa_opts, Config),
+            [{cert_key_alg, ecdsa} |
+             lists:delete(cert_key_alg,
+                          [{client_cert_opts, COpts},
+                           {server_cert_opts, SOpts} |
+                           lists:delete(server_cert_opts,
+                                        lists:delete(client_cert_opts, Config))]
+                         )];
+        false ->
+            {skip, "Missing EC crypto support"}
+    end.
+
 %%--------------------------------------------------------------------
 %% Test Cases --------------------------------------------------------
 %%--------------------------------------------------------------------
@@ -431,7 +492,7 @@ missing_root_cert_no_auth(Config) when is_list(Config) ->
 missing_root_cert_auth() ->
     [{doc,"Must have ROOT certs to be able to verify verify peer"}].
 missing_root_cert_auth(Config) when is_list(Config) ->
-    ServerOpts =  proplists:delete(cacertfile, ssl_test_lib:ssl_options(server_cert_opts, Config)),
+    ServerOpts =  proplists:delete(cacertfile, ssl_test_lib:ssl_options(extra_server, server_cert_opts, Config)),
     {ClientNode, ServerNode, _} = ssl_test_lib:run_where(Config),
     Version = proplists:get_value(version, Config),
     Server = ssl_test_lib:start_server_error([{node, ServerNode}, {port, 0},
@@ -441,7 +502,7 @@ missing_root_cert_auth(Config) when is_list(Config) ->
 
     ssl_test_lib:check_result(Server, {error, {options, {cacertfile, ""}}}),
     
-    ClientOpts =  proplists:delete(cacertfile, ssl_test_lib:ssl_options(client_cert_opts, Config)),
+    ClientOpts =  proplists:delete(cacertfile, ssl_test_lib:ssl_options(extra_client, client_cert_opts, Config)),
     Client = ssl_test_lib:start_client_error([{node, ClientNode}, {port, 0},
 					      {from, self()},
 					      {options, [{verify, verify_peer}
@@ -455,7 +516,7 @@ missing_root_cert_auth_user_verify_fun_accept() ->
      " with a verify_fun that accepts the unknown CA error"}].
 
 missing_root_cert_auth_user_verify_fun_accept(Config) ->
-    ServerOpts = ssl_test_lib:ssl_options(server_cert_opts, Config),
+    ServerOpts = ssl_test_lib:ssl_options(extra_server, server_cert_opts, Config),
     FunAndState =  {fun(_,{bad_cert, unknown_ca}, UserState) ->
 			    {valid, UserState};
 		       (_,{bad_cert, _} = Reason, _) ->
@@ -467,8 +528,8 @@ missing_root_cert_auth_user_verify_fun_accept(Config) ->
 		       (_, valid_peer, UserState) ->
 			    {valid, UserState}
 		    end, []},
-    ClientOpts = ssl_test_lib:ssl_options([{verify, verify_peer},
-                                            {verify_fun, FunAndState}], Config),
+    ClientOpts = ssl_test_lib:ssl_options(extra_client,  [{verify, verify_peer},
+                                                         {verify_fun, FunAndState}], Config),
     ssl_test_lib:basic_test(ClientOpts, ServerOpts, Config).
 
 %%--------------------------------------------------------------------
@@ -476,7 +537,7 @@ missing_root_cert_auth_user_backwardscompatibility_verify_fun_accept() ->
     [{doc, "Test old style verify fun"}].
 
 missing_root_cert_auth_user_backwardscompatibility_verify_fun_accept(Config) ->
-    ServerOpts = ssl_test_lib:ssl_options(server_cert_opts, Config),
+    ServerOpts = ssl_test_lib:ssl_options(extra_server, server_cert_opts, Config),
     AcceptBadCa = fun({bad_cert,unknown_ca}, Acc) ->  Acc;
                      (Other, Acc) -> [Other | Acc]
 		  end,
@@ -487,9 +548,8 @@ missing_root_cert_auth_user_backwardscompatibility_verify_fun_accept(Config) ->
 		    [_|_] -> false
 		end
 	end,
-    
-    ClientOpts = ssl_test_lib:ssl_options([{verify, verify_peer},
-                                           {verify_fun, VerifyFun}], Config),
+    ClientOpts = ssl_test_lib:ssl_options(extra_client, [{verify, verify_peer},
+                                                         {verify_fun, VerifyFun}], Config),
     ssl_test_lib:basic_test(ClientOpts, ServerOpts, Config).
 
 %%--------------------------------------------------------------------
@@ -498,7 +558,7 @@ missing_root_cert_auth_user_verify_fun_reject() ->
      " with a verify_fun that rejects the unknown CA error"}].
 
 missing_root_cert_auth_user_verify_fun_reject(Config) ->
-    ServerOpts = ssl_test_lib:ssl_options(server_cert_opts, Config),
+    ServerOpts = ssl_test_lib:ssl_options(extra_server, server_cert_opts, Config),
     FunAndState =  {fun(_,{bad_cert, unknown_ca} = Reason, _UserState) ->
 			    {fail, Reason};
 		       (_,{bad_cert, _} = Reason, _) ->
@@ -510,8 +570,8 @@ missing_root_cert_auth_user_verify_fun_reject(Config) ->
 		       (_, valid_peer, UserState) ->
 			    {valid, UserState}
 		    end, []},
-    ClientOpts =  ssl_test_lib:ssl_options([{verify, verify_peer},
-                                            {verify_fun, FunAndState}], Config),
+    ClientOpts =  ssl_test_lib:ssl_options(extra_client, [{verify, verify_peer},
+                                                          {verify_fun, FunAndState}], Config),
     ssl_test_lib:basic_alert(ClientOpts, ServerOpts, Config, unknown_ca).
 %%--------------------------------------------------------------------
 incomplete_chain_auth() ->
@@ -526,12 +586,12 @@ incomplete_chain_auth(Config) when is_list(Config) ->
                                                                          {client_chain, DefaultCertConf}]),   
     [ServerRoot| _] = ServerCas = proplists:get_value(cacerts, ServerOpts0),
     ClientCas = proplists:get_value(cacerts, ClientOpts0),
-    ClientOpts = ssl_test_lib:ssl_options([{verify, verify_peer},
-                                            {cacerts,  ServerCas ++ ClientCas} | 
-                                            proplists:delete(cacerts, ClientOpts0)], Config),
-    ServerOpts = ssl_test_lib:ssl_options([{verify, verify_peer},
-                                           {cacerts, [ServerRoot]} |  
-                                           proplists:delete(cacerts, ServerOpts0)], Config),
+    ClientOpts = ssl_test_lib:ssl_options(extra_client, [{verify, verify_peer},
+                                                         {cacerts,  ServerCas ++ ClientCas} |
+                                                         proplists:delete(cacerts, ClientOpts0)], Config),
+    ServerOpts = ssl_test_lib:ssl_options(extra_server, [{verify, verify_peer},
+                                                         {cacerts, [ServerRoot]} |
+                                                         proplists:delete(cacerts, ServerOpts0)], Config),
     ssl_test_lib:basic_test(ClientOpts, ServerOpts, Config).
 
 %%--------------------------------------------------------------------
@@ -540,8 +600,8 @@ verify_fun_always_run_client() ->
       "valid_peer not only unknown_extension)"}].
 
 verify_fun_always_run_client(Config) when is_list(Config) ->
-    ClientOpts =  ssl_test_lib:ssl_options(client_cert_opts, Config),
-    ServerOpts =  ssl_test_lib:ssl_options(server_cert_opts, Config),
+    ClientOpts =  ssl_test_lib:ssl_options(extra_client, client_cert_opts, Config),
+    ServerOpts =  ssl_test_lib:ssl_options(extra_server, server_cert_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
     Version = proplists:get_value(version, Config),
     Server = ssl_test_lib:start_server_error([{node, ServerNode}, {port, 0},
@@ -581,8 +641,8 @@ verify_fun_always_run_server() ->
     [{doc,"Verify that user verify_fun is always run (for valid and "
       "valid_peer not only unknown_extension)"}].
 verify_fun_always_run_server(Config) when is_list(Config) ->
-    ClientOpts =  ssl_test_lib:ssl_options(client_cert_opts, Config),
-    ServerOpts =  ssl_test_lib:ssl_options(server_cert_opts, Config),
+    ClientOpts =  ssl_test_lib:ssl_options(extra_client, client_cert_opts, Config),
+    ServerOpts =  ssl_test_lib:ssl_options(extra_server, server_cert_opts, Config),
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
 
     %% If user verify fun is called correctly we fail the connection.
@@ -644,8 +704,8 @@ critical_extension_auth(Config) when is_list(Config) ->
                                                                          [{server_chain, 
                                                                            [[],[],[{extensions, Ext}]]},
                                                                           {client_chain, DefaultCertConf}]),
-    ClientOpts = ssl_test_lib:ssl_options(ClientOpts0, Config),
-    ServerOpts = ssl_test_lib:ssl_options(ServerOpts0, Config),              
+    ClientOpts = ssl_test_lib:ssl_options(extra_client, ClientOpts0, Config),
+    ServerOpts = ssl_test_lib:ssl_options(extra_server, ServerOpts0, Config),
  
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
     Version = proplists:get_value(version, Config),
@@ -676,8 +736,8 @@ critical_extension_client_auth(Config) when is_list(Config) ->
                                                                          [{client_chain, 
                                                                            [[],[],[{extensions, Ext}]]},
                                                                           {server_chain, DefaultCertConf}]),
-    ClientOpts = ssl_test_lib:ssl_options(ClientOpts0, Config),
-    ServerOpts = ssl_test_lib:ssl_options(ServerOpts0, Config),              
+    ClientOpts = ssl_test_lib:ssl_options(extra_client, ClientOpts0, Config),
+    ServerOpts = ssl_test_lib:ssl_options(extra_server, ServerOpts0, Config),
     
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
     Version = proplists:get_value(version, Config),
@@ -711,8 +771,8 @@ critical_extension_no_auth(Config) when is_list(Config) ->
                                                                             [{server_chain, 
                                                                               [[],[], [{extensions, Ext}]]},
                                                                              {client_chain, DefaultCertConf}]),
-    ClientOpts = [{verify, verify_none} | ssl_test_lib:ssl_options(ClientOpts0, Config)],
-    ServerOpts = [{verify, verify_none} | ssl_test_lib:ssl_options(ServerOpts0, Config)],            
+    ClientOpts = [{verify, verify_none} | ssl_test_lib:ssl_options(extra_client, ClientOpts0, Config)],
+    ServerOpts = [{verify, verify_none} | ssl_test_lib:ssl_options(extra_server, ServerOpts0, Config)],
     
      ssl_test_lib:basic_test(ClientOpts, ServerOpts, Config).
 
@@ -732,8 +792,8 @@ extended_key_usage_auth(Config) when is_list(Config) ->
                                                                           [[],[], [{extensions, Ext}]]},
                                                                          {client_chain, DefaultCertConf}
                                                                         ]),
-    ClientOpts = ssl_test_lib:ssl_options(ClientOpts0, Config),
-    ServerOpts = ssl_test_lib:ssl_options(ServerOpts0, Config),                                                     
+    ClientOpts = ssl_test_lib:ssl_options(extra_client, ClientOpts0, Config),
+    ServerOpts = ssl_test_lib:ssl_options(extra_server, ServerOpts0, Config),
 
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
     Version = proplists:get_value(version, Config),
@@ -767,8 +827,8 @@ extended_key_usage_client_auth(Config) when is_list(Config) ->
       server_config := ServerOpts0} = ssl_test_lib:make_cert_chains_der(proplists:get_value(cert_key_alg, Config),
                                                                         [{client_chain, [[],[],[{extensions, ClientExt}]]},
                                                                          {server_chain, [[],[],[{extensions, ServerExt}]]}]),
-    ClientOpts = ssl_test_lib:ssl_options(ClientOpts0, Config),
-    ServerOpts = ssl_test_lib:ssl_options(ServerOpts0, Config),        
+    ClientOpts = ssl_test_lib:ssl_options(extra_client, ClientOpts0, Config),
+    ServerOpts = ssl_test_lib:ssl_options(extra_server, ServerOpts0, Config),
    
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
     Version = proplists:get_value(version, Config),
@@ -805,8 +865,8 @@ cert_expired(Config) when is_list(Config) ->
                                                                            []
                                                                      ]},
                                                                          {client_chain, DefaultCertConf}]),
-    ClientOpts = ssl_test_lib:ssl_options(ClientOpts0, Config),
-    ServerOpts = ssl_test_lib:ssl_options(ServerOpts0, Config),                                                     
+    ClientOpts = ssl_test_lib:ssl_options(extra_client, ClientOpts0, Config),
+    ServerOpts = ssl_test_lib:ssl_options(extra_server, ServerOpts0, Config),
     
     {ClientNode, ServerNode, Hostname} = ssl_test_lib:run_where(Config),
     Version = proplists:get_value(version, Config),
@@ -832,8 +892,8 @@ no_auth_key_identifier_ext(Config) when is_list(Config) ->
         ssl_test_lib:make_cert_chains_der(proplists:get_value(cert_key_alg, Config),
                                           [{client_chain, DefaultCertConf},
                                            {server_chain, DefaultCertConf}]),
-    ClientOpts = [{verify, verify_peer} | ssl_test_lib:ssl_options(ClientOpts0, Config)],
-    ServerOpts = [{verify, verify_peer} | ssl_test_lib:ssl_options(ServerOpts0, Config)],        
+    ClientOpts = [{verify, verify_peer} | ssl_test_lib:ssl_options(extra_client, ClientOpts0, Config)],
+    ServerOpts = [{verify, verify_peer} | ssl_test_lib:ssl_options(extra_server, ServerOpts0, Config)],
  
     ssl_test_lib:basic_test(ClientOpts, ServerOpts, Config).
     
@@ -852,8 +912,8 @@ no_auth_key_identifier_ext_keyEncipherment(Config) when is_list(Config) ->
                                             [[],[],[{extensions, ClientExt}]]},
                                            {server_chain, DefaultCertConf}
                                           ]),
-    ClientOpts =   [{verify, verify_peer} | ssl_test_lib:ssl_options(ClientOpts0, Config)],
-    ServerOpts =   [{verify, verify_peer} | ssl_test_lib:ssl_options(ServerOpts0, Config)],  
+    ClientOpts =   [{verify, verify_peer} | ssl_test_lib:ssl_options(extra_client, ClientOpts0, Config)],
+    ServerOpts =   [{verify, verify_peer} | ssl_test_lib:ssl_options(extra_server, ServerOpts0, Config)],
     
     ssl_test_lib:basic_test(ClientOpts, ServerOpts, Config).
 
@@ -872,10 +932,10 @@ key_auth_ext_sign_only(Config) when is_list(Config) ->
                                            {server_chain, DefaultCertConf}
                                           ]),
     Version = proplists:get_value(version, Config),
-    ClientOpts =  [{verify, verify_peer} | ssl_test_lib:ssl_options(ClientOpts0, Config)], 
+    ClientOpts =  [{verify, verify_peer} | ssl_test_lib:ssl_options(extra_client, ClientOpts0, Config)],
     ServerOpts = [{verify, verify_peer}, {ciphers, 
                                           ssl_test_lib:rsa_non_signed_suites(n_version(Version))} 
-                  | ssl_test_lib:ssl_options(ServerOpts0, Config)],
+                  | ssl_test_lib:ssl_options(extra_server, ServerOpts0, Config)],
     
     ssl_test_lib:basic_test(ClientOpts, ServerOpts, Config).
 
@@ -896,17 +956,17 @@ longer_chain(Config) when is_list(Config) ->
     [ServerRoot| _] = ServerCas = proplists:get_value(cacerts, ServerOpts0),
     ClientCas = proplists:get_value(cacerts, ClientOpts0),
     
-    ServerOpts = ssl_test_lib:ssl_options([{verify, verify_peer}, {cacerts, [ServerRoot]} |  
+    ServerOpts = ssl_test_lib:ssl_options(extra_server, [{verify, verify_peer}, {cacerts, [ServerRoot]} |
                                            proplists:delete(cacerts, ServerOpts0)], Config),
-    ClientOpts = ssl_test_lib:ssl_options([{verify, verify_peer},
+    ClientOpts = ssl_test_lib:ssl_options(extra_client, [{verify, verify_peer},
                                            {depth, 5},
                                            {cacerts,  ServerCas ++ ClientCas} | 
                                            proplists:delete(cacerts, ClientOpts0)], Config),
     ssl_test_lib:basic_test(ClientOpts, ServerOpts, Config).
 
-duplicate_chain() ->
-    [{doc, "Manual test of chain with duplicate entries"}].
-duplicate_chain(Config)
+cross_signed_chain() ->
+    [{doc, "Manual test of chain with duplicate entries that is cross signing certs are present."}].
+cross_signed_chain(Config)
   when is_list(Config) ->
     Key1 = ssl_test_lib:hardcode_rsa_key(1),
     Key2 = ssl_test_lib:hardcode_rsa_key(2),
@@ -935,30 +995,80 @@ duplicate_chain(Config)
     {[_Peer,CI1New,CI2New,CRNew], CRNew} = chain_and_root(ClientOptsNew),
 
     ServerCas = [CRNew|ServerCas0 -- [CROld]],
-    ServerOpts = ssl_test_lib:ssl_options([{verify, verify_peer} |
-                                           lists:keyreplace(cacerts, 1, ServerOpts0, {cacerts, ServerCas})],
+    ServerOpts = ssl_test_lib:ssl_options(extra_server, [{verify, verify_peer} |
+                                                         lists:keyreplace(cacerts, 1, ServerOpts0, {cacerts, ServerCas})],
                                           Config),
-    ClientOpts = ssl_test_lib:ssl_options([{verify, verify_peer} |
-                                           lists:keyreplace(cacerts, 1,
-                                                            lists:keyreplace(cert, 1, ClientOpts0,
+    ClientOpts = ssl_test_lib:ssl_options(extra_client, [{verify, verify_peer} |
+                                                         lists:keyreplace(cacerts, 1,
+                                                                          lists:keyreplace(cert, 1, ClientOpts0,
                                                                              {cert, [Peer,CI1New,CI2New,CI1,CI2,CRNew,CROld]}),
-                                                            {cacerts, ClientCas0})],
+                                                                          {cacerts, ClientCas0})],
                                           Config),
     ssl_test_lib:basic_test(ClientOpts, ServerOpts, Config),
-    ClientOpts2 = ssl_test_lib:ssl_options([{verify, verify_peer} |
-                                            lists:keyreplace(cacerts, 1,
-                                                             lists:keyreplace(cert, 1, ClientOpts0,
-                                                                              {cert, [Peer,CI1,CI1New,CI2,CI2New,CROld,CRNew]}),
-                                                             {cacerts, ClientCas0})],
+    ClientOpts2 = ssl_test_lib:ssl_options(extra_client, [{verify, verify_peer} |
+                                                          lists:keyreplace(cacerts, 1,
+                                                                           lists:keyreplace(cert, 1, ClientOpts0,
+                                                                                            {cert, [Peer,CI1,CI1New,CI2,CI2New,CROld,CRNew]}),
+                                                                           {cacerts, ClientCas0})],
                                            Config),
     ssl_test_lib:basic_test(ClientOpts2, ServerOpts, Config),
     ok.
 
-chain_and_root(Config) ->
-    OwnCert = proplists:get_value(cert, Config),
-    {ok, ExtractedCAs} = ssl_pkix_db:extract_trusted_certs({der, proplists:get_value(cacerts, Config)}),
-    {ok, Root, Chain} = ssl_certificate:certificate_chain(OwnCert, ets:new(foo, []), ExtractedCAs, []),
-    {Chain, Root}.
+expired_root_with_cross_signed_root() ->
+    [{doc,"Test that we can verify a chain with an expired Root Cert if there is an alternative chain with"
+      " a cross signed Root CA further down the chain that is however not present in the sent chain"}].
+expired_root_with_cross_signed_root(Config) when is_list(Config) ->
+
+    Key1 = ssl_test_lib:hardcode_rsa_key(1),
+    Key2 = ssl_test_lib:hardcode_rsa_key(2),
+    Key3 = ssl_test_lib:hardcode_rsa_key(3),
+    Key4 = ssl_test_lib:hardcode_rsa_key(4),
+    Key5 = ssl_test_lib:hardcode_rsa_key(5),
+    Key6 = ssl_test_lib:hardcode_rsa_key(6),
+    {Year, Month, Day} = date(),
+
+    %% Create expired ROOT
+    #{cert := Root} = SRoot = public_key:pkix_test_root_cert("OTP test server ROOT", [{key, Key1},
+                                                                    {validity, {{Year-2, Month, Day},
+                                                                                {Year-1, Month, Day}}}]),
+    #{server_config := ServerOpts, client_config := ClientOpts} =
+        public_key:pkix_test_data(#{server_chain => #{root => SRoot,
+                                                      intermediates => [[{key, Key2}], [{key, Key3}]],
+                                                      peer => [{key, Key4}]},
+                                    client_chain => #{root => [{key, Key5}],
+                                                      peer => [{key, Key6}]}}),
+
+    SCert = proplists:get_value(cert, ServerOpts),
+    SCerts = proplists:get_value(cacerts, ServerOpts),
+
+    {ok, ExtractedCAs} = ssl_pkix_db:extract_trusted_certs({der, SCerts}),
+    {ok, Root, [_Peer, CA1, CA2, Root]} = ssl_certificate:certificate_chain(SCert, ets:new(foo, []), 
+                                                                            ExtractedCAs, [], encoded),
+
+    OTPCA1 = public_key:pkix_decode_cert(CA1, otp),
+    OTPCA2 = public_key:pkix_decode_cert(CA2, otp),
+
+    TBS1 = OTPCA1#'OTPCertificate'.tbsCertificate,
+    TBS2 = OTPCA2#'OTPCertificate'.tbsCertificate,
+
+    Issuer = TBS1#'OTPTBSCertificate'.issuer,
+
+    SubjectPublicKeyInfo = TBS2#'OTPTBSCertificate'.subjectPublicKeyInfo,
+
+    AltCrossRoot = public_key:pkix_sign(TBS1#'OTPTBSCertificate'{subject = Issuer,
+                                                                 subjectPublicKeyInfo = SubjectPublicKeyInfo}, Key2),
+
+    ClientCas0 = proplists:get_value(cacerts, ClientOpts),
+
+    %% Only expired ROOT present
+    ssl_test_lib:basic_alert([{verify, verify_peer} | ClientOpts], ServerOpts, Config, certificate_expired),
+    %% Only CROSS ROOT present
+    ssl_test_lib:basic_test([{verify, verify_peer},
+                             {cacerts, [AltCrossRoot]} | proplists:delete(cacerts, ClientOpts)], ServerOpts, Config),
+    %% Both expired ROOT and CROSS ROOT present
+    ssl_test_lib:basic_test([{verify, verify_peer},
+                             {cacerts, [AltCrossRoot | ClientCas0]} | proplists:delete(cacerts, ClientOpts)],
+                            ServerOpts, Config).
 
 %%--------------------------------------------------------------------
 %% TLS 1.3 Test cases  -----------------------------------------------
@@ -996,33 +1106,35 @@ custom_groups(Config) ->
 %% of CertificateRequest does not contain the algorithm of the client certificate).
 %% ssl client sends an empty certificate.
 unsupported_sign_algo_cert_client_auth() ->
-     [{doc,"TLS 1.3: Test client authentication with unsupported signature_algorithm_cert"}].
+     [{doc,"TLS 1.3 (backported to TLS-1.2) : Test client authentication with unsupported signature_algorithm_cert"}].
 
 unsupported_sign_algo_cert_client_auth(Config) ->
-    ClientOpts0 = ssl_test_lib:ssl_options(client_cert_opts, Config),
+    ClientOpts = ssl_test_lib:ssl_options(client_cert_opts, Config),
     ServerOpts0 = ssl_test_lib:ssl_options(server_cert_opts, Config),
-    ServerOpts = [{versions, ['tlsv1.2','tlsv1.3']},
-                  {verify, verify_peer},
+    ServerOpts = [{verify, verify_peer},
                   {signature_algs, [rsa_pkcs1_sha256, rsa_pkcs1_sha384, rsa_pss_rsae_sha256, rsa_pss_pss_sha256]},
                   %% Skip rsa_pkcs1_sha256!
                   {signature_algs_cert, [rsa_pkcs1_sha384, rsa_pkcs1_sha512]},
                   {fail_if_no_peer_cert, true}|ServerOpts0],
-    ClientOpts = [{versions, ['tlsv1.2','tlsv1.3']}|ClientOpts0],
-    ssl_test_lib:basic_alert(ClientOpts, ServerOpts, Config, certificate_required).
+    Version = proplists:get_value(version, Config),
+    case Version of
+        'tlsv1.3' ->
+            ssl_test_lib:basic_alert(ClientOpts, ServerOpts, Config, certificate_required);
+        _  ->
+            ssl_test_lib:basic_alert(ClientOpts, ServerOpts, Config, insufficient_security)
+    end.
 
 %%--------------------------------------------------------------------
 unsupported_sign_algo_client_auth() ->
-     [{doc,"TLS 1.3: Test client authentication with unsupported signature_algorithm"}].
+     [{doc,"TLS 1.3 (backported to TLS-1.2): Test client authentication with unsupported signature_algorithm"}].
 
 unsupported_sign_algo_client_auth(Config) ->
     ClientOpts0 = ssl_test_lib:ssl_options(client_cert_opts, Config),
     ServerOpts0 = ssl_test_lib:ssl_options(server_cert_opts, Config),
-    ServerOpts = [{versions, ['tlsv1.2','tlsv1.3']},
-                  {verify, verify_peer},
-                  %% Skip rsa_pkcs1_sha256!
-                  {signature_algs, [rsa_pkcs1_sha384, rsa_pkcs1_sha512]},
+    ClientOpts = [{signature_algs, [rsa_pkcs1_sha256, rsa_pss_rsae_sha256]} | ClientOpts0],
+    ServerOpts = [{verify, verify_peer},
+                  {signature_algs, [ecdsa_sha1, rsa_pss_pss_sha256]},
                   {fail_if_no_peer_cert, true}|ServerOpts0],
-    ClientOpts = [{versions, ['tlsv1.2','tlsv1.3']}|ClientOpts0],
     ssl_test_lib:basic_alert(ClientOpts, ServerOpts, Config, insufficient_security).
 %%--------------------------------------------------------------------
 hello_retry_client_auth() ->
@@ -1081,6 +1193,57 @@ hello_retry_client_auth_empty_cert_rejected(Config) ->
    
     ssl_test_lib:basic_alert(ClientOpts, ServerOpts, Config, certificate_required).
 
+%%--------------------------------------------------------------------
+signature_algorithms_bad_curve_secp256r1() ->
+     [{doc,"TLS 1.3: Test that the the client fails to connect "
+       "if server's certificate has a key using an unsupported curve."}].
+
+signature_algorithms_bad_curve_secp256r1(Config) ->
+    ClientOpts0 = ssl_test_lib:ssl_options(client_cert_opts, Config),
+    ServerOpts0 = ssl_test_lib:ssl_options(server_cert_opts, Config),
+    %% Set versions
+    ServerOpts = [{versions, ['tlsv1.2','tlsv1.3']}, {log_level, debug}|ServerOpts0],
+    ClientOpts = [{versions, ['tlsv1.2','tlsv1.3']},
+                  {signature_algs, [ecdsa_secp384r1_sha384,
+                                    ecdsa_secp521r1_sha512,
+                                    {sha256,rsa}]}|ClientOpts0],
+
+    ssl_test_lib:basic_alert(ClientOpts, ServerOpts, Config, insufficient_security).
+
+%%--------------------------------------------------------------------
+signature_algorithms_bad_curve_secp384r1() ->
+     [{doc,"TLS 1.3: Test that the the client fails to connect "
+       "if server's certificate has a key using an unsupported curve."}].
+
+signature_algorithms_bad_curve_secp384r1(Config) ->
+    ClientOpts0 = ssl_test_lib:ssl_options(client_cert_opts, Config),
+    ServerOpts0 = ssl_test_lib:ssl_options(server_cert_opts, Config),
+    %% Set versions
+    ServerOpts = [{versions, ['tlsv1.2','tlsv1.3']}, {log_level, debug}|ServerOpts0],
+    ClientOpts = [{versions, ['tlsv1.2','tlsv1.3']},
+                  {signature_algs, [ecdsa_secp256r1_sha256,
+                                    ecdsa_secp521r1_sha512,
+                                    {sha256,rsa}]}|ClientOpts0],
+
+    ssl_test_lib:basic_alert(ClientOpts, ServerOpts, Config, insufficient_security).
+
+%%--------------------------------------------------------------------
+signature_algorithms_bad_curve_secp521r1() ->
+     [{doc,"TLS 1.3: Test that the the client fails to connect "
+       "if server's certificate has a key using an unsupported curve."}].
+
+signature_algorithms_bad_curve_secp521r1(Config) ->
+    ClientOpts0 = ssl_test_lib:ssl_options(client_cert_opts, Config),
+    ServerOpts0 = ssl_test_lib:ssl_options(server_cert_opts, Config),
+    %% Set versions
+    ServerOpts = [{versions, ['tlsv1.2','tlsv1.3']}, {log_level, debug}|ServerOpts0],
+    ClientOpts = [{versions, ['tlsv1.2','tlsv1.3']},
+                  {signature_algs, [ecdsa_secp256r1_sha256,
+                                    ecdsa_secp384r1_sha384,
+                                    {sha256,rsa}]}|ClientOpts0],
+
+    ssl_test_lib:basic_alert(ClientOpts, ServerOpts, Config, insufficient_security).
+
 %%--------------------------------------------------------------------
 basic_rsa_1024() ->
     [{doc, "TLS 1.3 (Basic): Test if connection can be established using 1024 bits RSA keys in certificates."}].
@@ -1119,3 +1282,18 @@ no_reuse({3, N}) when N >= 4 ->
     [];
 no_reuse(_) ->
     [{reuse_sessions, false}].
+
+chain_and_root(Config) ->
+    OwnCert = proplists:get_value(cert, Config),
+    {ok, ExtractedCAs} = ssl_pkix_db:extract_trusted_certs({der, proplists:get_value(cacerts, Config)}),
+    {ok, Root, Chain} = ssl_certificate:certificate_chain(OwnCert, ets:new(foo, []), ExtractedCAs, [], encoded),
+    {Chain, Root}.
+
+sig_algs(rsa_pss_pss) ->
+    [{signature_algs, [rsa_pss_pss_sha512,
+                       rsa_pss_pss_sha384,
+                       rsa_pss_pss_sha256]}];
+sig_algs(rsa_pss_rsae) ->
+    [{signature_algs, [rsa_pss_rsae_sha512,
+                       rsa_pss_rsae_sha384,
+                       rsa_pss_rsae_sha256]}].
